import numpy as np
import pandas as pd

import warnings
from pathlib import Path
import os,stat              # for chmod'ing output scripts and type-checking
import copy                 # for writing to file
from collections import abc # for type-checking
import contextlib           # for Runner to use multiple log files
import subprocess           # for Runner to execute rrtm from Python
import re                   # for Runner cleanup matching folder names
from abc import ABC, abstractmethod

from fortranformat import FortranRecordReader, FortranRecordWriter, config

class Runner:
    def __init__(self,n_workers=8,tmp_path='/private/tmp/lima',shell='/usr/local/bin/lima sh -c',
                 exec_path='~/linuxvm/rrtm_lw_pgf90/rrtm_v3.3_linux_pgf90',clean=True):
        #set attrs
        self.n_workers = n_workers
        self.clean     = clean
        self.tmp_path  = Path.expanduser(Path(tmp_path))
        self.exec_path = Path.expanduser(Path(exec_path))
        if shell:
            self.shell = shell
        else:
            self.shell = ''
        #pick unqiue identifier
        self.uid = '{:04d}'.format(np.random.randint(10000))
        #remove other worker directories
        if self.clean:
            Runner.rmdirs(self.tmp_path,'any')
        else:
            Runner.rmdirs(self.tmp_path,self.uid)
        #make worker directories
        dirs = Runner.mkdirs(self.n_workers,self.tmp_path,self.uid)
        self.dirs = dirs
        #link executable into worker directories
        Runner.lndirs(self.dirs,self.exec_path)
        return

    def mkdirs(n_workers,tmp_dir,uid):
        '''create worker directories'''
        dirpaths = [tmp_dir/f'rrtm{uid}_worker{i:02d}' for i in range(n_workers)]
        _        = [Path.mkdir(p) for p in dirpaths]
        print(f'created {len(dirpaths)} worker directories at {tmp_dir}')
        return dirpaths

    def lndirs(dirpaths,exec_path):
        '''symlink executable into worker dirs'''
        for p in dirpaths:
            subprocess.run(['ln','-s',(exec_path).resolve(),'rrtm'],cwd=p,check=True)
        return

    def rmdirs(path,uid,rms=['OUTPUT_RRTM','INPUT_RRTM','IN_CLD_RRTM',"'&1'",r'\&1',
                             'OUT_CLD_RRTM','TAPE6','TAPE7','rrtm.log','rrtm']):
        '''remove unused rrtm worker directories'''

        #if uid is specified, match that 4-digit string.
        #if uid == 'any', match any 4-digit number
        if uid=='any':
            uid='\\d{4}'

        #make a regex to match worker directory names
        #format is rrtmXXXX_workerYY: XXXX is uid and YY is worker number
        exp = re.compile(f'rrtm{uid}_worker\\d+')

        #get all matching paths in path
        paths = filter(lambda p: exp.fullmatch(p.name),path.glob('*'))
        paths = list(paths)

        #remove directories
        for p in paths:
            #remove all accepted temporary files
            _ = [(p/name).unlink(missing_ok=True) for name in rms]
            #try removing directory
            p.rmdir()
        print(f'removed {len(paths)} worker directories at {path}')
        return

    def run(self,inputs,verbose=True,ignored_warnings = ['Warning: ieee_underflow is signaling',
                                                         'Warning: ieee_inexact is signaling',
                                                         'FORTRAN STOP']):
        '''run RRTM on a list of parserrtm.Inputs and return output'''
        outputs = [ ]
        logs = [ ]
        i=0
        if verbose: print(f'starting {len(inputs)} jobs')
        while i<len(inputs):
            if verbose: print(f'{len(inputs)-i} jobs remaining')
            
            #wipe worker directories
            if verbose: print(f'\t wiping {len(self.dirs)} directories. Running:')
            for dir in self.dirs:
                p = subprocess.run(self.argproc('rm OUTPUT_RRTM INPUT_RRTM IN_CLD_RRTM OUT_CLD_RRTM TAPE6 TAPE7 rrtm.log'),
                                   cwd=dir,stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
            
            ps = [ ]
            logfilepaths = [dir/'rrtm.log' for dir in self.dirs]
            with contextlib.ExitStack() as stack:
                logfiles = [stack.enter_context(open(fname,'w')) for fname in logfilepaths]
                #submit jobs
                for j,dir in enumerate(self.dirs):
                    if i<len(inputs):
                        input = inputs[i]
                    else:
                        break
                    if verbose: print(f'\t \t {input.CXID[1:16]} in {dir.name}')
                    input.write(fpath=dir,file='auto')
                    p = subprocess.Popen(self.argproc('./rrtm'),shell=False,cwd=dir,stdout=logfiles[j],stderr=subprocess.STDOUT)
                    ps.append(p)
                    #print(f'\t \t {i}')
                    i+=1

                if verbose: print('\t waiting...',end='')
                #wait for jobs to finish and check for successful execution
                for j,p in enumerate(ps):
                    returncode = p.wait()
                    if returncode != 0:
                        print(f'worker {self.dirs[j]} failed!')
                        raise subprocess.CalledProcessError(returncode,p.args)
                if verbose: print('finished')
                #print(f'\t {i}')
        
            #collect outputs
            if verbose: print(f'\t reading {len(ps)} outputs')
            for p,dir in zip(ps,self.dirs):
                #read outputs
                outputs.append(read_output(dir/'OUTPUT_RRTM'))
                #read logs
                with open(dir/'rrtm.log','r') as f:
                    log=f.read().splitlines() 
                    #check for successful completion
                    if log[-1] != 'FORTRAN STOP':
                        raise UserError(f"RRTM run in {dir}/rrtm.log did not terminate sucessfully! 'FORTRAN STOP' expected, got {log[-1]} instead")
                    for line in log:
                        if line not in ignored_warnings:
                            warnings.warn(f'{dir}/rrtm.log has non-accepted warning \'{line}\'')
                    logs.append(log)
        if verbose: print(f'{len(inputs)-i} jobs remaining')
        if verbose: print(f'finished {len(inputs)} jobs')
        return outputs, logs

    def argproc(self,cmd):
        '''format arguments to subprocess based on self.shell'''
        return splitfirst(prepend(self.shell,cmd))

def prepend(shell,cmd):
    '''prepend shell to cmd if shell is non-empty str'''
    return [shell,cmd] if len(shell)>0 else [cmd]

def splitfirst(l):
    '''split first list element along spaces'''
    return [*l[0].split(),*(l[1:] if len(l)>1 else [])]

def read_output(fpath):
    calc_type = determine_output_type(fpath)
    if calc_type == 'longwave':
        return read_output_lw(fpath)
    elif calc_type == 'shortwave':
        return read_output_sw(fpath)

def read_output_lw(fpath):
    '''
    Read RRTM_LW ASCII output file and parse to xarray.Dataset.
    Adapted from https://github.com/climatedyn/RRTM_LW.
    Arguments:
        Absolute path to RRTM_LW output file
    Returns:
        Xarray.Dataset with contents of file
    '''
    
    #loop over file to get number of numeric lines to read
    with open(fpath,'r') as f:
        #load lines as list
        lnes = f.readlines()
        #find the line # of a form-feed+newline (this marks the end of the table)
        indx = lnes.index('\x0c\n')
        #get lines before column names and after table to save as metadata
        global_attrs = {'source': fpath,
                        'created_by':'parserrtm',
                        'output_header': lnes[0],
                        'output_footer': lnes[indx+1:]}

    #parse with pandas
    ds = pd.read_fwf(
                        fpath,
                        header=[1,2], #2nd and 3rd lines (1st gets auto-skipped)
                        nrows=indx-3, #remove three leading text lines
                        widths=[8,13,14,14,14,19], #field widths
                        index_col=0 #use LEVEL for index
                    ).to_xarray()

    #put units in attrs, format varnames
    ds = ds.rename({'index':'level'})
    rname = { }
    for var in ds.data_vars:
        if var[0].replace(' ','_').lower() == 'netflux': 
            rname[var] = 'net_flux'
        else:
            rname[var] = var[0].replace(' ','_').lower()
        ds[var].attrs['units'] = var[1]
    ds = ds.rename(rname)
    #store non-table data in global attrs
    ds.attrs = global_attrs
    return ds

def read_output_sw(fpath):
    '''
    Read RRTM_SW ASCII output file and parse to xarray.Dataset.
    Arguments:
        Absolute path to RRTM_SW output file
    Returns:
        Xarray.Dataset with contents of file
    '''
    #loop over file to get number of numeric lines to read
    with open(fpath,'r') as f:
        #load lines as list
        lines = f.readlines()
        #find the first line of the table (after "Wavenumbers: ...")
        for i,line in enumerate(lines):
            if line[1:12] == 'Wavenumbers':
                start_i = i+1
                break
        #find the line # of a form-feed+newline (this marks the end of the table)
        end_i = lines.index('\x0c\n')
        #get lines before column names and after table to save as metadata
        global_attrs = {'source': fpath,
                        'created_by':'parserrtm',
                        'output_header': lines[0],
                        'output_footer': lines[end_i+1:]}

        #parse with pandas
        ds = pd.read_fwf(
                            fpath,
                            header=[start_i-1, start_i], #2nd and 3rd lines (1st gets auto-skipped)
                            nrows=end_i-start_i-2, #remove three leading text lines
                            widths=[6,10,14,14,15,14,15,13], #field widths
                            index_col=0 #use LEVEL for index
                        ).to_xarray()

        #put units in attrs, format varnames
        ds = ds.rename({'index':'level'})
        rname = { }
        for var in ds.data_vars:
            if var[0].replace(' ','_').lower() == 'netflux': 
                rname[var] = 'net_flux'
            else:
                rname[var] = var[0].replace(' ','_').lower()
            ds[var].attrs['units'] = var[1]
        ds = ds.rename(rname)
        #store non-table data in global attrs
        ds.attrs = global_attrs
        return ds

def determine_output_type(fpath):
    with open(fpath) as f:
        i = 0
        while i<3:
            line = f.readline()
            i+=1
            #print(line[1:34])
            if line[1:34] == 'Wavenumbers:   10.0 - 3250.0 cm-1':
                return 'longwave'
            elif line[1:34] == 'Wavenumbers:   820. - 50000. cm-1':
                return 'shortwave'




    
from .input import Input
import warnings

class InputSW(Input):
    '''Specification of inputs for an RRTM-SW calculation

    RRTM has many options which require different input variables.
    This class stores input variables and automates the I/O process
    for the complex text files required to run RRTM-SW.
    
    See the file `rrtm_instructions` as a part of the RRTM-SW source code
    for definitions of each variable name.

    See the examples distributed as a part of this repository for examples
    of the variables necessary for various kinds of calculations.
    '''

    def read_input_rrtm(self, fpath):
        '''
        Read and interpret "INPUT_RRTM" text file into the current instance.
        
        "INPUT_RRTM" contains the overall model setup and specifies surface
        boundary conditions, temperature and pressure profiles, which gases
        to simulate, and their concentration at levels.
        
        Scan the content of a file and interpret each line according to 
        the format, sequence, and naming described in the documentation
        file "rrtm_instructions". Each field is read into an attribute
        of the current instance. Repeated fields (e.g. user-defined profiles) 
        are read into a list in order of occurrence.

        -------------
        Arguments:
        self        : instance of parserrtm.Input
        fpath       : path to "INPUT_RRTM" file
        -------------
        Returns:
        self        : instance with all fields stored as attributes
        -------------
        '''
        #1. get file record layout
        #2. get field lists
        #3. loop over lines and store
        
        with open(fpath,'r') as f:
            #read lines into list
            lines = f.readlines()
            self.lines = lines
            
        #get start and end lines
        start_i, end_i = Input.get_input_rrtm_file_bounds(self.lines)
        
        #set read position to starting line
        self.read_i = start_i
        
        #sequentially read records of file
        self.read_record('1.1')
        self.read_record('1.2')
        self.read_record('1.2.1')
        self.read_record('1.4')
        if self.IATM == 0:
            self.read_record('2.1')
            for i in range(self.NLAYRS):
                self.read_record('2.1.1',mode='new' if i==0 else 'append')
                self.read_record('2.1.2',mode='new' if i==0 else 'append')
                if self.NMOL > 7:
                    self.read_record('2.1.3',mode='new' if i==0 else 'append')
        elif self.IATM == 1:
            self.read_record('3.1')
            self.read_record('3.2')
            if self.IBMAX == 0:
                self.read_record('3.3A')
            else:
                self.read_greedy_record('3.3B') #GREEDY
            if self.MODEL == 0:
                self.read_record('3.4')
                for i in range(abs(self.IMMAX)):
                    self.read_record('3.5',mode='new' if i==0 else 'append')
                    self.read_record('3.6.1',mode='new' if i==0 else 'append')
            # Records 3.8 to 3.8.2 are stated to be present in rrtm_sw_instructions,
            # but this is inconsistent.
            # if self.IPRFL == 0: 
            #     self.read_record('3.8')
            #     for i in range(self.LAYX):
            #         self.read_record('3.8.1',mode='new' if i==0 else 'append')
            #         self.read_record('3.8.2',mode='new' if i==0 else 'append')
                    
        if self.read_i != end_i:
            warnings.warn(f'{fpath} read finished on line {self.read_i} instead of {end_i} -- some of input is unread!')

        #determine derived parameter NSTR (# of mom. of phase function needed)
        # this is used for determining the format of both in_cld_rrtm and in_aer_rrtm
        self.NSTR = 0
        if self.ISTRM == 0:
            self.NSTR = 4
        elif self.ISTRM == 1:
            self.NSTR = 8
        elif self.ISTRM == 2:
            self.NSTR = 16
        else:
            raise ValueError(f"'{self.ISTRM}' not a valid ISTRM value")
            
        del self.lines
        return self

    def read_in_cld_rrtm(self, fpath):
        '''
        Read and interpret "IN_CLD_RRTM" text file into the current instance.
        
        "IN_CLD_RRTM" contains information about cloud layers in the model and
        is only used if ICLD (record 1.2) = 1 or 2. Note that an "INPUT_RRTM"
        file must already be read into self before reading "IN_CLD_RRTM",
        since the formatting of "IN_CLD_RRTM" depends on some fields from
        "INPUT_RRTM" (namely ISCAT and NUMANGS).
        
        Scan the content of a file and interpret each line according to 
        the format, sequence, and naming described in the documentation
        file "rrtm_instructions". Each field is read into an attribute
        of the current instance. Repeated fields (e.g. user-defined profiles) 
        are read into a list in order of occurrence.

        -------------
        Arguments:
        self        : instance of rrtmparse Input class
        fpath       : path to "IN_CLD_RRTM" file
        -------------
        Returns:
        self        : instance with all fields stored as attributes
        -------------
        '''
        
        with open(fpath,'r') as f:
            #read lines into list
            lines = f.readlines()
            self.lines = lines
            
        #get start and end lines (start line is zero)
        start_i, end_i = Input.get_input_rrtm_file_bounds(self.lines,file='in_cld_rrtm')
        
        #set read position to starting line
        self.read_i = start_i
        
        #iterate/read over records
        self.read_record('C1.1')
        if self.INFLAG == 0:
            self.read_record('C1.2',mode='new')
            while self.read_i < end_i:
                self.read_record('C1.2',mode='append')
        elif self.INFLAG == 2:
            self.read_record('C1.3',mode='new')
            while self.read_i < end_i:
                self.read_record('C1.3',mode='append')
        
        #check that we read up to the expected end of records
        if self.read_i != end_i:
            warnings.warn(f'{fpath} read finished on line {self.read_i} instead of {end_i} -- some of input is unread!')
            
        del self.lines
        return self


    def get_logical_record_order(self,file='input_rrtm'):
            '''
            Get record order to print out. Logical order means
            no repeated records.
            
            Options: file = 'input_rrtm', 'in_cld_rrtm', 'in_aer_rrtm'
            '''
            
            #go through logic
            if file == 'input_rrtm':
                records = ['1.1','1.2','1.2.1','1.4']
                if self.IATM == 0:
                    records.append('2.1')
                    records.append('2.1.1')
                    records.append('2.1.2')
                    if self.NMOL > 7:
                        records.append('2.1.3')
                elif self.IATM == 1:
                    records.append('3.1')
                    records.append('3.2')
                    if self.IBMAX == 0:
                        records.append('3.3A')
                    else:
                        records.append('3.3B') #GREEDY
                    if self.MODEL == 0:
                        records.append('3.4')
                        records.append('3.5')
                        records.append('3.6.1')
                    # if self.IPRFL == 0: 
                    #     records.append('3.8')
                    #     records.append('3.8.1')
                    #     records.append('3.8.2')

            elif file == 'in_cld_rrtm':
                #iterate/read over records
                records = ['C1.1']
                if self.INFLAG == 0:
                        records.append('C1.2')
                elif self.INFLAG == 2:
                        records.append('C1.3')

            elif file == 'in_aer_rrtm':
                records = ['A1.1']
                records.append('A2.1')
                records.append('A2.1.1')
                for i in range(self.NAER):
                    records.append('A2.2')
                    records.append('A2.3')
            
            return records
    
    def get_explicit_record_order(self,file='input_rrtm'):
        '''Get line-by-line records to print out.
        
        Options: file = 'input_rrtm', 'in_cld_rrtm', 'in_aer_rrtm'
        '''
        #go through logic
        if file == 'input_rrtm':
            records = ['1.1','1.2','1.2.1','1.4']
            if self.IATM == 0:
                records.append('2.1')
                for i in range(self.NLAYRS):
                    records.append('2.1.1')
                    records.append('2.1.2')
                    if self.NMOL > 7:
                        records.append('2.1.3')
            elif self.IATM == 1:
                records.append('3.1')
                records.append('3.2')
                if self.IBMAX == 0:
                    records.append('3.3A')
                else:
                    records.append('3.3B') #GREEDY
                if self.MODEL == 0:
                    records.append('3.4')
                    for i in range(abs(self.IMMAX)):
                        records.append('3.5')
                        records.append('3.6.1')
                # if self.IPRFL == 0: 
                #     records.append('3.8')
                #     for i in range(self.LAYX):
                #         records.append('3.8.1')
                #         records.append('3.8.2')

        elif file == 'in_cld_rrtm':
            #iterate/read over records
            records = ['C1.1']
            if self.INFLAG == 0:
                for i in range(1 if type(self.CLAY) != list else len(self.CLAY)): #number of cloudy layers
                    records.append('C1.2')
            elif self.INFLAG == 2:
                for i in range(1 if type(self.CLAY) != list else len(self.CLAY)): #number of cloudy layers
                    records.append('C1.3')

        elif file == 'in_aer_rrtm':
                records = ['A1.1']
                for i in range(self.NAER):
                    records.append('A2.1')
                    for j in range(self.NLAY[-1] if isinstance(self.NLAY,list) else self.NLAY):
                        records.append('A2.1.1')
                    records.append('A2.2')
                    records.append('A2.3')
                    if (self.IPHA[-1] if isinstance(self.IPHA,list) else self.IPHA) == 2:
                        raise NotImplementedError('Full phase-function specification not supported!')
        return records
    
    def get_format(self,rec): 
        '''
        Get dict of Fortran format-strings for each record.

        It's a method since some record formats (e.g., xsec records 2.1.1-3 and 2.2.4-5) 
        depend on the values of other records (IFORM and IFRMX).
        '''
        
        #dictionary of all record formats. Dynamic formats are stored as methods.
        #changed all E formats to ES and two E formats to F to align with example input file encoding
        formats = {
            '1.1':    '(1A80)',
            '1.2':    '(18X, I2, 29X, I1, 32X, I1, 1X, I1, 2X, I3, 4X, I1, 3X, I1, I1)',
            #RRTM_SW v2.7.5: '(12X, I3, 3X, F7.4, 4X, I1,14F7.5)'
            #RRTM_SW v2.5: '(12X, I3, F7.4, 4X, I1, 14F5.3)'
            '1.2.1':  '(12X, I3, 3X, F7.4, 4X, I1,14F7.5)',
            '1.4':    '(11X,  I1, 2X, I1, 14F5.3)',
            '2.1':    '(1X,I1, I3, I5)', #same as LW, conditional on IATM=0
            '2.1.1':  lambda self: f"({'F10.4' if self.IFORM==0 else 'ES15.7'}, F10.4, 23X, F8.3, F7.2,  7X, F8.3,   F7.2)", #same as LW
            '2.1.2':  lambda self: f"({'8ES10.3' if self.IFORM==0 else '8ES15.7'})", #same as LW
            '2.1.3':  lambda self: f"({'8ES10.3' if self.IFORM==0 else '8ES15.7'})", #same as LW
            '3.1':    '(I5, 5X, I5, 5X, I5, I5, I5, 3X, I2, F10.3, 20X, F10.3, F10.3)',
            '3.2':    '(F10.3,  F10.3)', #same as LW
            '3.3A':   '(F10.3,  F10.3,  F10.3, F10.3, F10.3)', #same as LW
            '3.3B':   '(8F10.3)', #same as LW
            '3.4':    '(I5, 3A8)', #same as LW
            '3.5':    '(F10.3, F10.3, F10.3, 5X, A1, A1, 3X, 28A1)', #changed from E to F, same as LW
            '3.6.1':  '(8ES10.3)', #same as LW
            '3.8':    '(I5, I5, A50)', #same as LW
            '3.8.1':  '(F10.3, 5X, 35A1)', #same as LW
            '3.8.2':  '(8ES10.3)', #same as LW
            'C1.1':   '(4X, I1,  4X, I1,  4X, I1)', #very similar to LW
            'C1.2':   '(A1, 1X, I3, ES10.3, ES10.3, ES10.3, 16ES10.3)', #changed from E10.5 to ES10.3
            'C1.3':   '(A1, 1X, I3, ES10.3, ES10.3, ES10.3, ES10.3, ES10.3)', #changed from E10.5 to ES10.3, same as LW
            'A1.1':   '(3X, I2)',
            'A2.1':   '(3X, I2, 4X, I1, 4X, I1, 4X, I1, 3F8.2)',
            'A2.1.1': lambda self: f"(2X, I3, {'F7.4' if self.IAOD == 0 else '14F7.4'})",
            'A2.2':   '(14F5.2)',
            'A2.3':   '(14F5.2)'
        }
        
        #get format
        fmt = formats[rec]
        
        #if format is a method, evaluate it
        if hasattr(fmt, '__call__'):
            fmt = fmt(self)
            
        #return format
        return fmt
    
    def get_fields(self,rec):
        '''
        Get dict of lists of field names for each record.

        It's a method since the number and names of many records' fields 
        depend on a field stored in some previous record. If a value is a
        function, it will be evaluated with the fields currently read in
        as attributes.
        '''
        
        #dictionary of all record formats. Dynamic formats are stored as methods.
        records_fields = {
            '1.1':     ['CXID'],
            '1.2':     ['IAER', 'IATM', 'ISCAT',  'ISTRM',  'IOUT', 'ICLD', 'IDELM', 'ICOS'], #different from LW
            '1.2.1':   ['JULDAT', 'SZA', 'ISOLVAR', *[f'SOLVAR({IB})' for IB in range(16,29+1)]], #different from LW
            '1.4':     ['IEMIS', 'IREFLECT', *[f'SEMISS({IB})' for IB in range(16,29+1)]], #different from LW
            '2.1':     ['IFORM', 'NLAYRS', 'NMOL'],
            '2.1.1':   ['PAVE',  'TAVE',    'PZ(L-1)',  'TZ(L-1)',   'PZ(L)',  'TZ(L)'],
            '2.1.2':   ['WKL(1,L)','WKL(2,L)','WKL(3,L)','WKL(4,L)','WKL(5,L)','WKL(6,L)','WKL(7,L)','WBROAD(L)'],
            '2.1.3':   lambda self: [f'WKL({M},L)' for M in range(self.NMOL-7)],
            '3.1':     ['MODEL',   'IBMAX',  'NOPRNT',  'NMOL', 'IPUNCH',   'MUNITS',    'RE',      'CO2MX', 'REF_LAT'], #different from LW
            '3.2':     ['HBOUND','HTOA'],
            '3.3A':    ['AVTRAT', 'TDIFF1', 'TDIFF2', 'ALTD1', 'ALTD2'],
            '3.3B':    lambda self: [f"{'Z' if self.IBMAX>0 else 'P'}BND({I})" for I in range(1, abs(self.IBMAX)+1)],
            '3.4':     ['IMMAX','HMOD'],
            '3.5':     ['ZM', 'PM', 'TM', 'JCHARP', 'JCHART', *[f'JCHAR({K})'for K in range(1,28+1)]],
            '3.6.1':   lambda self: [f'VMOL({K})'for K in range(1,self.NMOL+1)],
            '3.8':     ['LAYX','IZORP','XTITLE'],
            '3.8.1':   ['ZORP', *[f'JCHARX({K})'for K in range(1,28+1)]], #JCHAR(K) is already taken by record 3.5
            '3.8.2':   lambda self: [f'DENX({K})' for K in range(1, self.IXMOLS+1)],
            'C1.1':    ['INFLAG', 'ICEFLAG', 'LIQFLAG'],
            'C1.3':    lambda self: ['TESTCHAR','CLAY','CLDFRAC', 'TAUCLD' if self.INFLAG==0 else 'CWP',
                        'FRACICE','EFFSIZEICE','EFFSIZELIQ'], #C1.3 and C1.2 are swapped from LW
            'C1.2':    lambda self: ['TESTCHAR', 'CLAY', 'CLDFRAC', 'TAUCLD' if self.INFLAG==0 else 'CWP', #changed LAY to CLAY to differentiate from aerosol
                                     'SINGLE-SCATTERING ALBEDO', *[f'PMOM({N})' for N in range(self.NSTR)]],
            'A1.1':    ['NAER'],
            'A2.1':    ['NLAY', 'IAOD', 'ISSA', 'IPHA', 'AERPAR(1)', 'AERPAR(2)', 'AERPAR(3)'],
            'A2.1.1': lambda self: ['ALAY','AOD1'] if self.IAOD == 0 else ['ALAY', *[f'AOD({IB})' for IB in range(16,29+1)]], #changed LAY to ALAY to differentiate from cloud
            'A2.2':    [*[f'SSA({IB})' for IB in range(16,29+1)]],
            'A2.3':    [*[f'PHASE({IB})' for IB in range(16,29+1)]]
        }
        
        #get fields
        fields = records_fields[rec]
        
        #if fields is a method, evaluate it
        if hasattr(fields, '__call__'):
            fields = fields(self)
            
        #return fields
        return fields


import numpy as np
import pandas as pd

import warnings
from pathlib import Path
import os,stat              # for chmod'ing output scripts and type-checking
import copy                 # for writing to file
from collections import abc # for type-checking
import contextlib           # for Runner to use multiple log files
import subprocess           # for Runner to execute rrtm from Python
import re                   # for Runner cleanup matching folder names
from abc import ABC, abstractmethod

from fortranformat import FortranRecordReader, FortranRecordWriter, config


config.RET_UNWRITTEN_VARS_NONE = False #zero-fill instead of NA-fill missing values

def fpaths_union(fpaths_a, fpaths_b, verbose=True):
    fna, fnb = [p.name for p in fpaths_a], [p.name for p in fpaths_b]
    fnab = set(fna).intersection(fnb)
    fpaths_a_thin = filter(lambda p: p.name in fnab, fpaths_a)
    fpaths_b_thin = filter(lambda p: p.name in fnab, fpaths_b)
    if verbose:
        a_but_not_b = set(fna).difference(fnb)
        b_but_not_a = set(fnb).difference(fna)
        if a_but_not_b:
            print(f'removed {a_but_not_b} from fpaths_a')
        if b_but_not_a:
            print(f'remove {b_but_not_a} from fpaths_b')
    return list(fpaths_a_thin), list(fpaths_b_thin)

def read_input(*args):
    if len(args) == 1:
        return Input(args[0])
    else:
        return Input(args)

class Input(ABC):
    '''
    Abstract base class for representing RRTM Inputs

    This allows for code sharing between longwave and shortwave variants, 
    which have different formats for the necessary input files.
    '''
    def __repr__(self):
        recs = { }
        for rec in ['1.1','1.2','1.4']:
            for field in self.get_fields(rec):
                val = getattr(self,field)
                recs[field] = val
        return str(recs)
    
    def __str__(self):
        recs = { }
        for rec in ['1.1','1.2','1.4']:
            for field in self.get_fields(rec):
                val = getattr(self,field)
                recs[field] = val
        return str(recs)
    
    def __init__(self, args, **kwargs):
        
        #treat as a filename if argument is filepath-like
        if isinstance(args, (str, os.PathLike)):
            fpath = args
            self.read_input_rrtm(fpath, **kwargs)
            self.fpath_input_rrtm = fpath

        #treat as a dictionary of fields if argument is dict-like
        elif isinstance(args, abc.Mapping):
            self.from_dict(args,**kwargs)

        
        
        #treat as a list of filenames if argument is list-like
        # if length 1, assume input_rrtm
        elif isinstance(args, abc.Sequence) and (len(args)==1):
            fpath = args
            self.read_input_rrtm(fpath[0],**kwargs)
            self.fpath_input_rrtm = fpath[0]

        # if length 2, assume input_rrtm and in_cld_rrtm
        elif isinstance(args, abc.Sequence) and (len(args)==2):
            fpath = args
            self.read_input_rrtm(fpath[0],**kwargs)
            self.read_in_cld_rrtm(fpath[1],**kwargs)
            self.fpath_input_rrtm = fpath[0]
            self.fpath_in_cld_rrtm = fpath[1]

        # if length 3, assume input_rrtm, in_cld_rrtm, and in_aer_rrtm
        # allow for in_cld_rrtm to be None or 'none'
        elif isinstance(args, abc.Sequence) and (len(args)==3):
            fpath = args
            self.read_input_rrtm(fpath[0],**kwargs)
            self.fpath_input_rrtm = fpath[0]
            if fpath[1]:
                if str(fpath[1]).lower() != 'none':
                    self.read_in_cld_rrtm(fpath[1],**kwargs)
                    self.fpath_in_cld_rrtm = fpath[1]
            self.read_in_aer_rrtm(fpath[2],**kwargs)
            self.fpath_in_aer_rrtm = fpath[2]
        
        else:
            raise TypeError('Expected one of {} or {} of length 2 or 3, got {} instead'.format(
                (abc.Mapping,str,os.PathLike),abc.Sequence, type(args)))
        
        #derived parameter for number of streams,
        #specified differently for LW and SW.
        #only needed for cloudy cases
        if self.ICLD == 1:
            self.NSTR = 0
            #shortwave
            if hasattr(self,'ISTRM'):
                if self.ISTRM == 0:
                    self.NSTR = 4
                elif self.ISTRM == 1:
                    self.NSTR = 8
                elif self.ISTRM == 2:
                    self.NSTR = 16
            #longwave
            if hasattr(self,'NUMANGS'):
                if (self.ISCAT==1) or (self.ISCAT==2):
                    if self.NUMANGS==0:
                        self.NSTR=4
                    elif self.NUMANGS==1:
                        self.NSTR=8
                    elif self.NUMANGS==2:
                        self.NSTR=16
                    else:
                        raise ValueError(f"'{self.NUMANGS}' not a valid NUMANGS value, \
                        must be 0, 1, or 2 when ISCAT > 0")
        
    def __getitem__(self, item):
         return getattr(self,item)

    def __setitem__(self, item, val):
        setattr(self,item, val)

    def from_dict(self,args,file='cld',lazy=False):
        '''Add dictionary keys to a parserrtm.Input object.
        
        Dictionary keys are treated as field names. Optionally
        initialize needed fields for valid file(s).
        
        Parameters
        ---------
        self : parserrtm.Input
            Input object to which to apply dictionary
        args : dict-like
            dictionary containing field:value pairs to apply
        file : str, optional
            key indicating which file types for which to fill in missing fields.
            Options 'gas','aer','cld','aercld'. Default is 'cld'. Meanings:
            - gas:    input_rrtm only
            - aer:    input_rrtm and in_aer_rrtm
            - cld:    input_rrtm and in_cld_rrtm
            - aercld: input_rrtm, in_cld_rrtm, and in_aer_rrtm
            Only used if lazy is False.
        lazy : bool, optional
            Skip checking and filling in any needed fields for a valid input file 
            that are missing in args. Default is False (i.e. filling-in is the default).
    
        Returns
        --------
        self : parserrtm.Input
            Input object with fields initialized according to arguments
        '''
        #1. put all supplied fields into self
        for name,value in args.items():
            setattr(self,name,value)
    
        #2. initialize any missing fields with default values, if lazy==False
        if not lazy:
            # initialize fields needed to write different files based on "file" argument
            
            # always prepare input_rrtm
            #1. fill in fields needed for logical record order,
            #   if not already supplied. NB: this list is excessive
            required = ['IATM','ICLD','IAER','NMOL','IXSECT','IXMOLS','IBMAX','MODEL','IXSECT','IPRFL']
            for name in required:
                if not hasattr(self,name):
                    default = 7 if name == 'NMOL' else 0
                    setattr(self,name,default)
        
            #2. get all needed fields via logical record order
            records = self.get_logical_record_order('input_rrtm')
            for record in records:
                names = self.get_fields(record)
                for name in names:
                    if not hasattr(self,name):
                        if 'CHAR' in name:
                            default = " "
                        else:
                            default = 0
                        setattr(self,name,default)
                            
            if file in ('cld','aercld'):
                # prepare in_cld_rrtm
                #1. fill in fields needed for logical record order,
                #   if not already supplied. 
                required = ['INFLAG']
                for name in required:
                    if not hasattr(self,name):
                        default = 0
                        setattr(self,name,default)
            
                #2. get all needed fields via logical record order
                records = self.get_logical_record_order('in_cld_rrtm')
                for record in records:
                    names = self.get_fields(record)
                    for name in names:
                        if not hasattr(self,name):
                            if 'CHAR' in name:
                                default = " "
                            else:
                                default = 0
                            setattr(self,name,default)

            #TODO: defaults and requirements for in_aer_rrtm are unknown
            if file in ('aer','aercld'):
                # prepare in_aer_rrtm
                #1. fill in fields needed for logical record order,
                #   if not already supplied. 
                required = ['INFLAG']
                for name in required:
                    if not hasattr(self,name):
                        default = 0
                        setattr(self,name,default)
            
                #2. get all needed fields via logical record order
                records = self.get_logical_record_order('in_aer_rrtm')
                for record in records:
                    names = self.get_fields(record)
                    for name in names:
                        if not hasattr(self,name):
                            if 'CHAR' in name:
                                default = " "
                            else:
                                default = 0
                            setattr(self,name,default)
            return self
        
    def broadcast_scalars(self,file='cld'):
        '''
        Broadcast any scalar fields in repeated (listed) records to proper lengths

        file : str, optional
            key indicating which file types for which to fill in missing fields.
            Options 'gas','aer','cld','aercld'. Default is 'cld'. Meanings:
            - gas:    input_rrtm only
            - aer:    input_rrtm and in_aer_rrtm
            - cld:    input_rrtm and in_cld_rrtm
            - aercld: input_rrtm, in_cld_rrtm, and in_aer_rrtm
        '''
        if file == 'cld':
            files = ('input_rrtm','in_cld_rrtm')
        elif file == 'gas':
            files = ('input_rrtm')
        elif file == 'aer':
            files = ('input_rrtm','in_aer_rrtm')
        elif file == 'aercld':
            files == ('input_rrtm','in_aer_rrtm')
        else:
            files = [file]
        for file in files:
            records = self.get_logical_record_order(file)
            for record in records:
                if Input.islist(record):
                    length = self.record_len(record)
                    names  = self.get_fields(record)
                    for name in names:
                        val = getattr(self,name)
                        if not isinstance(val,list):
                            setattr(self,name,[val]*length)
        return self
    
    def print(self):
        self.fancy_print('input_rrtm')
        if self.ICLD>0:
            self.fancy_print('in_cld_rrtm')
        if self.IAER>0:
            self.fancy_print('in_aer_rrtm')

    def fancy_print(self, file='input_rrtm'):
        '''
        Print out all fields record-by-record to the console
        '''
        
        records = self.get_logical_record_order(file)
        for rec in records:
            print('-----------------------------')
            print(f'{rec}: \n',end='')
            fields = self.get_fields(rec)
            d      = {key:getattr(self,key) for key in fields}
            if not Input.islist(rec):
                print(d)
            else:
                #dict of lists to DataFrame
                df = pd.DataFrame(d)
                with pd.option_context('display.max_rows', 10, 
                                       'display.max_columns', 10, 
                                       'display.float_format','{:,.2E}'.format):
                    print(df)
        print('-----------------------------')

    #TODO: verify that these lists are comprehensive between the two input specifications
    def islist(rec):
        '''
        return if record is a list (True) or scalar (False)
        '''
        lists = ['2.1.1','2.1.2','2.1.3','2.2.3','2.2.4','2.2.5','3.5','3.6.1',
                 '3.8.1','3.8.2','C1.2','C1.3','C1.3a','A2.1','A2.1.1','A2.2','A2.3']
        return rec in lists

    def record_len(self,rec):
        '''
        Return number of times a record is repeated.
        '''
        lens = {
            '2.1.1': lambda self: self.NLAYRS,
            '2.1.2': lambda self: self.NLAYRS,
            '2.1.3': lambda self: self.NLAYRS,
            '2.2.3': lambda self: self.NLAYRS,
            '2.2.4': lambda self: self.NLAYRS,
            '2.2.5': lambda self: self.NLAYRS,
            '3.5':   lambda self: self.IMMAX,
            '3.6.1': lambda self: self.IMMAX,
            '3.8.1': lambda self: self.LAYX,
            '3.8.2': lambda self: self.LAYX,
            'C1.2':  lambda self: 1 if type(self.CLAY) != list else len(self.CLAY),
            'C1.3':  lambda self: 1 if type(self.CLAY) != list else len(self.CLAY),
            'C1.3a': lambda self: 15,
            'A2.1':  lambda self: self.NAER,
            'A2.1.1':lambda self: sum(self.NLAY),
            'A2.2':  lambda self: self.NAER,
            'A2.3':  lambda self: self.NAER if self.IPHA<2 else self.NSTR*self.NAER
        }
        return lens[rec](self)
    
    @abstractmethod
    def get_logical_record_order(self,file='input_rrtm'):
        pass

    @abstractmethod
    def get_explicit_record_order(self,file='input_rrtm'):
        pass

    def copy(self):
        return copy.deepcopy(self)

    def write(self,fpath=None, file='auto', fnames=None):
        '''
        Write instance of parserrtm.Input to text files needed to run RRTM.

        Note that these files are always called INPUT_RRTM, IN_CLD_RRTM, IN_AER_RRTM because this
        is required to run RRTM. Use folder organization for more descriptive filenames.
        
        -------------
        Arguments:
        self        : parserrtm.Input
        fpath       : (str or PathLike, optional) folder where input files are written
                      If fpath is not specified, files are written to the current working directory.
        file        : (str, optional) kind of file to write. Options are 'input_rrtm', 'in_cld_rrtm', 'in_aer_rrtm', or 'auto' (default).
        fname       : (list, optional) list of filenames to write. Specified in the order INPUT_RRTM, IN_CLD_RRTM, IN_AER_RRTM as
                      applicable according to `file`. By default, filenames are INPUT_RRTM, IN_CLD_RRTM, IN_AER_RRTM since this is what 
                      is required by the RRTM executable.

        -------------
        Returns:
        None
        -------------
        '''
        files = [ ]
        if file == 'auto':
            files.append('input_rrtm')
            if self.ICLD > 0:
                files.append('in_cld_rrtm')
            if hasattr(self,'IAER'):
                if self.IAER > 0:
                    files.append('in_aer_rrtm')

        # add directory
        if fpath == None:
            fpath = Path.cwd()
        fpaths = {'input_rrtm':  Path(fpath)/'INPUT_RRTM',
                'in_cld_rrtm': Path(fpath)/'IN_CLD_RRTM',
                'in_aer_rrtm': Path(fpath)/'IN_AER_RRTM'}
        
        # apply filenames
        if fnames:
            for i,file in enumerate(files):
                fpaths[file] = Path(fpath)/fnames[i]
        
        # write applicable files
        methods = {'input_rrtm':self.write_input_rrtm, 
                   'in_cld_rrtm':self.write_in_cld_rrtm, 
                   'in_aer_rrtm':self.write_in_aer_rrtm}
        for file in files:
            methods[file](fpaths[file])
            
        return
    
    def write_input_rrtm(self,fpath):
        records = self.get_explicit_record_order('input_rrtm')
        rundupe = self.copy()
        with open(fpath,'w') as f:
            f.write('\n')
            f.write('\n')
            f.write('file auto-generated by parserrtm\n')
            for rec in records:
                line = rundupe.write_record(rec)
                f.write(line+'\n')
            f.write('%')

    def write_in_cld_rrtm(self,fpath):
        records = self.get_explicit_record_order('in_cld_rrtm')
        rundupe = self.copy()
        with open(fpath,'w') as f:
            for rec in records:
                line = rundupe.write_record(rec)
                f.write(line+'\n')
            f.write('%\n')
            f.write('file auto-generated by parserrtm')

    def write_in_aer_rrtm(self,fpath):
        records = self.get_explicit_record_order('in_aer_rrtm')
        rundupe = self.copy()
        with open(fpath,'w') as f:
            for rec in records:
                line = rundupe.write_record(rec)
                f.write(line+'\n')
            f.write('%\n')
            f.write('file auto-generated by parserrtm')

    def write_record(self,rec):
        '''
        RRTM record writer. Parse a record line using Fortran formats.
        Destructively writes out lists (should be run on a copy of self).
        
        -------------
        Arguments:
        self        : instance of parserrtm.Input
        rec         : (str) record name to write out
        -------------
        Returns:
        line        : (str) line formatted to fields
        -------------
        '''
        #get params for reader by polling self
        fmt    = self.get_format(rec)
        names  = self.get_fields(rec)
        
        #get list of fields to write
        vals = [ ]
        for key in names:
            attr = getattr(self,key)
            if type(attr) == list:
                val = attr.pop(0)
            else:
                val = attr
            vals.append(val)
        
        #parse fields from file line
        writer = FortranRecordWriter(fmt)
        line   = writer.write(vals)
        
        return line
    
    @abstractmethod
    def read_in_cld_rrtm(self,fpath):
        pass

    @abstractmethod
    def read_input_rrtm(self,fpath):
        pass

    def read_in_aer_rrtm(self,fpath):
        '''
        Read and interpret "IN_AER_RRTM" text file into the current instance.
        
        "IN_AER_RRTM" contains information about aerosol layers in the model and
        is only used if IAER (record 1.2) = 10 in RRTM_SW. Note that an "INPUT_RRTM"
        file must already be read into self before reading "IN_AER_RRTM".
        ''' 
        with open(fpath,'r') as f:
            #read lines into list
            lines = f.readlines()
            self.lines = lines
            
        #get start and end lines (start line is zero)
        start_i, end_i = Input.get_input_rrtm_file_bounds(self.lines,file='in_aer_rrtm')
        
        #set read position to starting line
        self.read_i = start_i
        
        #iterate/read over records
        self.read_record('A1.1')
        for i in range(self.NAER):
            self.read_record('A2.1',mode='new' if i==0 else 'append')
            #NOTE: this approach gives AOD flattened dimensions of (0,...,NLAY(NAER=0),...,NLAY(NAER=0)+NLAY(NAER=1),...,sum(NLAY(NAER)))
            for j in range(self.NLAY[-1] if isinstance(self.NLAY,list) else self.NLAY):
                self.read_record('A2.1.1',mode='new' if (i == 0) & (j ==0) else 'append')
        for i in range(self.NAER):
            #fields are SSA(IB) with a length of NAER
            self.read_record('A2.2',mode='new' if i==0 else 'append')
            if (self.IPHA[-1] if isinstance(self.IPHA,list) else self.IPHA) == 2:
                for i in self.NSTR:
                    #left unsupported because records would be different.
                    #instead of IPHASE(IB) with dimensions of NAER,
                    # it would be IPHASE(IB,NSTR) with dimensions of NAER,
                    # where NSTR is an index for each stream.
                    raise NotImplementedError('Direct aerosol phase-function specification not yet supported!')
                    self.read_record('A2.3',mode='new' if i==0 else 'append')
            else:
                self.read_record('A2.3')
        
        #check that we read up to the expected end of records
        if self.read_i != end_i:
            warnings.warn(f'{fpath} read finished on line {self.read_i} instead of {end_i} -- some of input is unread!')
            
        del self.lines
        return self

    def get_input_rrtm_file_bounds(lines,file='input_rrtm'):
        '''
        Find start and end lines for "INPUT_RRTM"  or "IN_CLD_RRTM" file.
        
        -------------
        Arguments:
        lines (list): list of lines of file (from file.readlines())
        file   (str): type of input file ('input_rrtm', 'in_cld_rrtm', or 'in_aer_rrtm')
        -------------
        Returns:
        start_i, end_i (int): line positions of first and last lines of file
        -------------
        '''

        #find start ('$') and end ('%') positions
        starts = [i for i,s in enumerate(lines) if s[0]=='$']
        ends   = [i for i,s in enumerate(lines) if s[0]=='%']
        
        #in_cld_rrtm has no starting '$' character and begins on the first line
        if file in ['in_cld_rrtm','in_aer_rrtm']:
            starts = [0]

        #in_aer_rrtm has no ending '$' character required so assume input ends with file
        if file == 'in_aer_rrtm':
            ends.append(len(lines)-1)

        #check '$' and '%' only occur once and '$' comes before '%'
        if (len(starts)==len(ends)==1) and (starts<ends):
            start_i, end_i = starts[0], ends[0]
        else:
            raise IOError(f"start lines '{starts}' and end lines '{ends}' not valid. Start is a line beginning with '$',\
            end is a line beginning with '%', and input file must have exactly one start and end with the start occurring before the end.")

        return start_i, end_i
    
    def read_record(self,rec,mode='new'):
        '''
        RRTM record reader. Parse a line using Fortran formats.
        Read into self.
        
        -------------
        Arguments:
        self        : instance of RRTM_LW parser class
        rec         : which record next line of file is
        mode        : options ('new','append','append depth'): 
                            -- new: create/overwrite attribute with read value
                            -- append: append new value to end of what's already there
                            -- append depth: convert last element to list and append (create list of lists)
        -------------
        Returns:
        self        : instance with fields of record added as attributes
        -------------
        '''
        #get params for reader by polling self
        s      = self.lines[self.read_i]
        fmt    = self.get_format(rec)
        names  = self.get_fields(rec)
        
        #parse fields from file line
        reader = FortranRecordReader(fmt)
        fields = reader.read(s)
        
        #assign fields as attributes
        for key,val in zip(names,fields):
            if mode=='new':
                setattr(self,key,val)
            
            elif mode=='append':
                l = getattr(self,key)
                if type(l) != list:
                    l = [l]
                l.append(val)
                setattr(self,key,l)
                
            elif mode=='append depth':
                l  = getattr(self,key)
                l2 = l[-1]
                if type(l2) != list:
                    l2 = [l2]
                l2.append(val)
                l[-1] = l2
                setattr(self,key,l)
            
        #step forwards one line
        self.read_i += 1
        return
    
    def read_greedy_record(self,rec):
        '''
        RRTM record reader. Parse a line using Fortran formats.
        Read into self. Read lines until all fields of record are filled.
        Note this only works for lines with generic formats (e.g. '(8F10.3)').

        -------------
        Arguments:
        self        : instance of RRTM_LW parser class
        rec         : which record next line of file is
        -------------
        Returns:
        self        : instance with fields added as attributes
        -------------
        '''
        #get params for reader by polling self
        s      = self.lines[self.read_i]
        fmt    = self.get_format(rec)
        names  = self.get_fields(rec)
        
        #record desired number of fields
        inames = len(names)
        
        #make an empty record
        record = { }
        
        #read once
        #parse fields from file line
        reader = FortranRecordReader(fmt)
        fields = reader.read(s)
        for key,val in zip(names,fields):
            record[key] = val
        self.read_i += 1
                
        #if fields are still missing, keep reading
        while len(record.keys()) < inames:
            #remove keys from names
            names = list(filter(lambda var: var not in list(record.keys()),names))
            
            #read with remaining names
            #parse fields from file line
            fields = reader.read(self.lines[self.read_i])
            for key,val in zip(names,fields):
                record[key] = val
            self.read_i += 1
        
        #write back to self after finished
        for key,val in record.items():
            setattr(self,key,val)
        return
    
    @abstractmethod
    def get_format(self,rec): 
        '''
        Get dict of Fortran format-strings for each record.

        It's a method since some record formats (xsec records 2.1.1-3 and 2.2.4-5) 
        depend on the values of other records (IFORM and IFRMX).
        '''
        pass

    @abstractmethod
    def get_fields(self,rec):
        '''
        Get dict of lists of field names for each record.

        It's a method since the number and names of many records' fields 
        depend on a field stored in some previous record. If a value is a
        function, it will be evaluated with the fields currently read in
        as attributes.
        '''
        pass


    
from .input import Input
import warnings

class InputLW(Input):
    '''Specification of inputs for an RRTM-LW calculation

    RRTM has many options which require different input variables.
    This class stores input variables and automates the I/O process
    for the complex text files required to run RRTM-LW.
    
    See the file `rrtm_instructions` as a part of the RRTM-LW source code
    for definitions of each variable name.

    See the examples distributed as a part of this repository for examples
    of the variables necessary for various kinds of calculations.
    '''

    def read_input_rrtm(self, fpath):
        '''
        Read and interpret "INPUT_RRTM" text file into the current instance.
        
        "INPUT_RRTM" contains the overall model setup and specifies surface
        boundary conditions, temperature and pressure profiles, which gases
        to simulate, and their concentration at levels.
        
        Scan the content of a file and interpret each line according to 
        the format, sequence, and naming described in the documentation
        file "rrtm_instructions". Each field is read into an attribute
        of the current instance. Repeated fields (e.g. user-defined profiles) 
        are read into a list in order of occurrence.

        -------------
        Arguments:
        self        : instance of parserrtm.Input
        fpath       : path to "INPUT_RRTM" file
        -------------
        Returns:
        self        : instance with all fields stored as attributes
        -------------
        '''
        #1. get file record layout
        #2. get field lists
        #3. loop over lines and store
        
        with open(fpath,'r') as f:
            #read lines into list
            lines = f.readlines()
            self.lines = lines
            
        #get start and end lines
        start_i, end_i = Input.get_input_rrtm_file_bounds(self.lines)
        
        #set read position to starting line
        self.read_i = start_i
        
        #sequentially read records of file
        self.read_record('1.1')
        self.read_record('1.2')
        self.read_record('1.4')
        if self.IATM == 0:
            self.read_record('2.1')
            for i in range(self.NLAYRS):
                self.read_record('2.1.1',mode='new' if i==0 else 'append')
                self.read_record('2.1.2',mode='new' if i==0 else 'append')
                if self.NMOL > 7:
                    self.read_record('2.1.3',mode='new' if i==0 else 'append')
            if self.IXSECT == 1:
                self.read_record('2.2')
                self.read_record('2.2.1')
                self.read_record('2.2.2')
                for i in range(self.NLAYRS):
                    self.read_record('2.2.3',mode='new' if i==0 else 'append') #dummy
                    self.read_record('2.2.4',mode='new' if i==0 else 'append')
                    if self.IXMOLS > 7:
                        self.read_record('2.2.5',mode='new' if i==0 else 'append')
        elif self.IATM == 1:
            self.read_record('3.1')
            self.read_record('3.2')
            if self.IBMAX == 0:
                self.read_record('3.3A')
            else:
                self.read_greedy_record('3.3B') #GREEDY
            if self.MODEL == 0:
                self.read_record('3.4')
                for i in range(abs(self.IMMAX)):
                    self.read_record('3.5',mode='new' if i==0 else 'append')
                    self.read_record('3.6.1',mode='new' if i==0 else 'append')
            if self.IXSECT == 1:
                self.read_record('3.7')
                self.read_record('3.7.1')
                if self.IPRFL == 0: 
                    self.read_record('3.8')
                    for i in range(self.LAYX):
                        self.read_record('3.8.1',mode='new' if i==0 else 'append')
                        self.read_record('3.8.2',mode='new' if i==0 else 'append')
                        
        if self.read_i != end_i:
            warnings.warn(f'{fpath} read finished on line {self.read_i} instead of {end_i} -- some of input is unread!')
            
        del self.lines
        return self

    def read_in_cld_rrtm(self, fpath):
        '''
        Read and interpret "IN_CLD_RRTM" text file into the current instance.
        
        "IN_CLD_RRTM" contains information about cloud layers in the model and
        is only used if ICLD (record 1.2) = 1 or 2. Note that an "INPUT_RRTM"
        file must already be read into self before reading "IN_CLD_RRTM",
        since the formatting of "IN_CLD_RRTM" depends on some fields from
        "INPUT_RRTM" (namely ISCAT and NUMANGS).
        
        Scan the content of a file and interpret each line according to 
        the format, sequence, and naming described in the documentation
        file "rrtm_instructions". Each field is read into an attribute
        of the current instance. Repeated fields (e.g. user-defined profiles) 
        are read into a list in order of occurrence.
        
        NOTE: 
            For INFLAG=10, the fields TAUCLD, SINGLE-SCATTERING ALBEDO, PMOM(0:NSTR)
            are a list of lists. What this means: each item in the outer list is a 
            cloudy layer, while each item in the inner list are the cloud properties 
            for the 16 spectral bands in ascending order. 

            The first band is specified as C1.3, while the next 15 bands are 
            specified as C1.3a, which has fewer fields, which makes the formatting a bit
            unusual. Since no examples in the source code use INFLAG=10, this is untested!

        -------------
        Arguments:
        self        : instance of rrtmparse Input class
        fpath       : path to "IN_CLD_RRTM" file
        -------------
        Returns:
        self        : instance with all fields stored as attributes
        -------------
        '''
        
        with open(fpath,'r') as f:
            #read lines into list
            lines = f.readlines()
            self.lines = lines
            
        #get start and end lines (start line is zero)
        start_i, end_i = Input.get_input_rrtm_file_bounds(self.lines,file='in_cld_rrtm')
        
        #set read position to starting line
        self.read_i = start_i
        
        #determine derived parameter NSTR (# of mom. of phase function needed)
        self.NSTR = 0
        if (self.ISCAT==1) or (self.ISCAT==2):
            if self.NUMANGS==0:
                self.NSTR=4
            elif self.NUMANGS==1:
                self.NSTR=8
            elif self.NUMANGS==2:
                self.NSTR=16
            else:
                raise ValueError(f"'{self.NUMANGS}' not a valid NUMANGS value, \
                must be 0, 1, or 2 when ISCAT > 0")
                
        #iterate/read over records
        self.read_record('C1.1')
        if (self.INFLAG == 1) or (self.INFLAG == 2):
            self.read_record('C1.2',mode='new')
            while self.read_i < end_i:
                self.read_record('C1.2',mode='append')
        elif (self.INFLAG == 0) or (self.INFLAG == 10):
            self.read_record('C1.3',mode='new')
            while self.read_i < end_i:
                self.read_record('C1.3',mode='append')
                if self.INFLAG == 10:
                    for i in range(15):
                        self.read_record('C1.3a',mode='append depth')
        
        #check that we read up to the expected end of records
        if self.read_i != end_i:
            warnings.warn(f'{fpath} read finished on line {self.read_i} instead of {end_i} -- some of input is unread!')
            
        del self.lines
        return self

    def get_logical_record_order(self,file='input_rrtm'):
            '''
            Get record order to print out. Logical order means
            no repeated records.
            
            Options: file = 'input_rrtm', 'in_cld_rrtm'
            '''
            
            #go through logic
            if file == 'input_rrtm':
                records = ['1.1','1.2','1.4']
                if self.IATM == 0:
                    records.append('2.1')
                    records.append('2.1.1')
                    records.append('2.1.2')
                    if self.NMOL > 7:
                        records.append('2.1.3')
                    if self.IXSECT == 1:
                        records.append('2.2')
                        records.append('2.2.1')
                        records.append('2.2.2')
                        records.append('2.2.3') #dummy
                        records.append('2.2.4')
                        if self.IXMOLS > 7:
                            records.append('2.2.5')
                elif self.IATM == 1:
                    records.append('3.1')
                    records.append('3.2')
                    if self.IBMAX == 0:
                        records.append('3.3A')
                    else:
                        records.append('3.3B') #GREEDY
                    if self.MODEL == 0:
                        records.append('3.4')
                        records.append('3.5')
                        records.append('3.6.1')
                    if self.IXSECT == 1:
                        records.append('3.7')
                        records.append('3.7.1')
                        if self.IPRFL == 0: 
                            records.append('3.8')
                            records.append('3.8.1')
                            records.append('3.8.2')

            elif file == 'in_cld_rrtm':
                #iterate/read over records
                records = ['C1.1']
                if (self.INFLAG == 1) or (self.INFLAG == 2):
                        records.append('C1.2')
                elif (self.INFLAG == 0) or (self.INFLAG == 10):
                    for i in range(len(self.CLAY)): #number of cloudy layers
                        records.append('C1.3')
                        if self.INFLAG == 10:
                            raise NotImplementedError('INFLAG=10 (spectrally-resolved COD) not yet supported!')
                            for i in range(15):
                                records.append('C1.3a')
            
            return records
    
    def get_explicit_record_order(self,file='input_rrtm'):
        '''Get line-by-line records to print out.
        
        Options: file = 'input_rrtm', 'in_cld_rrtm'
        '''
        #go through logic
        if file == 'input_rrtm':
            records = ['1.1','1.2','1.4']
            if self.IATM == 0:
                records.append('2.1')
                for i in range(self.NLAYRS):
                    records.append('2.1.1')
                    records.append('2.1.2')
                    if self.NMOL > 7:
                        records.append('2.1.3')
                if self.IXSECT == 1:
                    records.append('2.2')
                    records.append('2.2.1')
                    records.append('2.2.2')
                    for i in range(self.NLAYRS):
                        records.append('2.2.3') #dummy
                        records.append('2.2.4')
                        if self.IXMOLS > 7:
                            records.append('2.2.5')
            elif self.IATM == 1:
                records.append('3.1')
                records.append('3.2')
                if self.IBMAX == 0:
                    records.append('3.3A')
                else:
                    records.append('3.3B') #GREEDY
                if self.MODEL == 0:
                    records.append('3.4')
                    for i in range(abs(self.IMMAX)):
                        records.append('3.5')
                        records.append('3.6.1')
                if self.IXSECT == 1:
                    records.append('3.7')
                    records.append('3.7.1')
                    if self.IPRFL == 0: 
                        records.append('3.8')
                        for i in range(self.LAYX):
                            records.append('3.8.1')
                            records.append('3.8.2')

        elif file == 'in_cld_rrtm':
            #iterate/read over records
            records = ['C1.1']
            if (self.INFLAG == 1) or (self.INFLAG == 2):
                for i in range(1 if type(self.CLAY) != list else len(self.CLAY)): #number of cloudy layers
                    records.append('C1.2')
            elif (self.INFLAG == 0) or (self.INFLAG == 10):
                for i in range(1 if type(self.CLAY) != list else len(self.CLAY)): #number of cloudy layers
                    records.append('C1.3')
                    if self.INFLAG == 10:
                        raise NotImplementedError('INFLAG=10 (spectrally-resolved COD) not yet supported!')
                        for i in range(15):
                            records.append('C1.3a')
        return records
    
    def get_format(self,rec): 
        '''
        Get dict of Fortran format-strings for each record.

        It's a method since some record formats (xsec records 2.1.1-3 and 2.2.4-5) 
        depend on the values of other records (IFORM and IFRMX).
        '''
        
        #dictionary of all record formats. Dynamic formats are stored as methods.
        #changed all E formats to ES and two E formats to F to align with example input file encoding
        formats = {
            '1.1':    '(1A80)',
            '1.2':    '(49X, I1, 19X, I1, 12X, I1, I2, 2X, I3, 4X, I1)',
            '1.4':    '(F10.3,  1X, I1, 2X, I1, 16F5.3)', #changed from E to F
            '2.1':    '(1X,I1, I3, I5)',
            '2.1.1':  lambda self: f"({'F10.4' if self.IFORM==0 else 'ES15.7'}, F10.4, 23X, F8.3, F7.2,  7X, F8.3,   F7.2)",
            '2.1.2':  lambda self: f"({'8ES10.3' if self.IFORM==0 else '8ES15.7'})",
            '2.1.3':  lambda self: f"({'8ES10.3' if self.IFORM==0 else '8ES15.7'})",
            '2.2':    '(I5)',
            '2.2.1':  '(7A10,(/,8A10))',
            '2.2.2':  '(1X,I1)',
            '2.2.3':  '(A1)', #dummy record
            '2.2.4':  lambda self: f"({'7ES10.3' if self.IFRMX==0 else '7ES15.7'})",
            '2.2.5':  lambda self: f"({'8ES10.3' if self.IFRMX==0 else '8ES15.7'})",
            '3.1':    '(I5, 5X, I5, 5X, I5, I5, I5, 3X, I2, F10.3, 20X, F10.3)',
            '3.2':    '(F10.3,  F10.3)',
            '3.3A':   '(F10.3,  F10.3,  F10.3, F10.3, F10.3)',
            '3.3B':   '(8F10.3)',
            '3.4':    '(I5, 3A8)',
            '3.5':    '(F10.3, F10.3, F10.3, 5X, A1, A1, 3X, 28A1)', #changed from E to F
            '3.6.1':  '(8ES10.3)',
            '3.7':    '(I5, I5, I5)',
            '3.7.1':  '(7A10,(/,8A10))',
            '3.8':    '(I5, I5, A50)',
            '3.8.1':  '(F10.3, 5X, 35A1)',
            '3.8.2':  '(8ES10.3)',
            'C1.1':   '(3X, I2,  4X, I1,  4X, I1)',
            'C1.2':   '(A1, 1X, I3, ES10.3, ES10.3, ES10.3, ES10.3, ES10.3)', #changed from E10.5 to ES10.3
            'C1.3':   '(A1, 1X, I3, ES10.3, ES10.3, ES10.3, ES10.3, ES10.3)', #changed from E10.5 to ES10.3
            'C1.3A':  '(15X, E10.5, E10.5, 16E10.5)'
        }
        
        #get format
        fmt = formats[rec]
        
        #if format is a method, evaluate it
        if hasattr(fmt, '__call__'):
            fmt = fmt(self)
            
        #return format
        return fmt
    
    def get_fields(self,rec):
        '''
        Get dict of lists of field names for each record.

        It's a method since the number and names of many records' fields 
        depend on a field stored in some previous record. If a value is a
        function, it will be evaluated with the fields currently read in
        as attributes.
        '''
        
        #dictionary of all record formats. Dynamic formats are stored as methods.
        records_fields = {
            '1.1':     ['CXID'],
            '1.2':     ['IATM',  'IXSECT', 'ISCAT',  'NUMANGS',   'IOUT',   'ICLD'],
            '1.4':     ['TBOUND',  'IEMIS', 'IREFLECT', *[f'SEMISS({IB})' for IB in range(1,16+1)]],
            '2.1':     ['IFORM', 'NLAYRS', 'NMOL'],
            '2.1.1':   ['PAVE',  'TAVE',    'PZ(L-1)',  'TZ(L-1)',   'PZ(L)',  'TZ(L)'],
            '2.1.2':   ['WKL(1,L)','WKL(2,L)','WKL(3,L)','WKL(4,L)','WKL(5,L)','WKL(6,L)','WKL(7,L)','WBROAD(L)'],
            '2.1.3':   lambda self: [f'WKL({M},L)' for M in range(self.NMOL-7)],
            '2.2':     ['IXMOLS'],
            '2.2.1':   lambda self: [f'XSNAME({I})' for I in range(1, self.IXMOLS+1)],
            '2.2.2':   ['IFRMX'],
            '2.2.3':   ['_'], #dummy record
            '2.2.4':   lambda self: [f'XAMNT({I})' for I in range(1, 8)],
            '3.1':     ['MODEL',   'IBMAX',  'NOPRNT',  'NMOL', 'IPUNCH',   'MUNITS',    'RE',      'CO2MX'],
            '3.2':     ['HBOUND','HTOA'],
            '3.3A':    ['AVTRAT', 'TDIFF1', 'TDIFF2', 'ALTD1', 'ALTD2'],
            '3.3B':    lambda self: [f"{'Z' if self.IBMAX>0 else 'P'}BND({I})" for I in range(1, abs(self.IBMAX)+1)],
            '3.4':     ['IMMAX','HMOD'],
            '3.5':     ['ZM', 'PM', 'TM', 'JCHARP', 'JCHART', *[f'JCHAR({K})'for K in range(1,28+1)]],
            '3.6.1':   lambda self: [f'VMOL({K})'for K in range(1,self.NMOL+1)],
            '3.7':     ['IXMOLS','IPRFL','IXSBIN'],
            '3.7.1':   lambda self: [f'XSNAME({I})' for I in range(1, self.IXMOLS+1)],
            '3.8':     ['LAYX','IZORP','XTITLE'],
            '3.8.1':   ['ZORP', *[f'JCHARX({K})'for K in range(1,28+1)]], #JCHAR(K) is already taken by record 3.5
            '3.8.2':   lambda self: [f'DENX({K})' for K in range(1, self.IXMOLS+1)],
            'C1.1':    ['INFLAG', 'ICEFLAG', 'LIQFLAG'],
            'C1.2':    ['TESTCHAR','CLAY','CLDFRAC','CWP','FRACICE','EFFSIZEICE','EFFSIZELIQ'], #changed LAY to CLAY to differentiate from aerosol
            'C1.3':    lambda self: ['TESTCHAR', 'CLAY', 'CLDFRAC', 'TAUCLD', 
                                     'SINGLE-SCATTERING ALBEDO', *[f'PMOM({N})' for N in range(self.NSTR)]],
            'C1.3A':   lambda self: ['TAUCLD', 'SINGLE-SCATTERING ALBEDO', *[f'PMOM({N})' for N in range(self.NSTR)]]
        }
        
        #get fields
        fields = records_fields[rec]
        
        #if fields is a method, evaluate it
        if hasattr(fields, '__call__'):
            fields = fields(self)
            
        #return fields
        return fields